"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var Fs = require("fs-extra");
var Tmp = require("tmp");
var local_storage_1 = require("./local-storage");
var paths_1 = require("./paths");
var utils_1 = require("./utils");
/**
 Contains general git utilities.
 */
var exec = utils_1.Utils.exec;
// This RegExp will help us pluck the versions in a conflict and solve it
var conflict = /\n\s*<<<<<<< [^\n]+(\n(?:.|\n)+?)\n\s*=======(\n(?:.|\n)+?)\n\s*>>>>>>> [^\n]+/;
function git(argv, options) {
    return gitBody(utils_1.Utils.git, argv, options);
}
var gitPrint = git.print = function (argv, options) {
    if (options === void 0) { options = {}; }
    return gitBody(utils_1.Utils.git.print, argv, options);
};
// Create a compare link e.g. https://github.com/Urigo/WhatsApp/compare/xxxxxxx..xxxxxxx
function reviewTutorial(remote, branch) {
    var remoteUrl = exports.Git(['remote', 'get-url', remote]);
    var remoteData = utils_1.pluckRemoteData(remoteUrl);
    if (process.env.NODE_ENV === 'test') {
        remoteData = { host: 'host', owner: 'owner', repo: 'repo' };
    }
    if (!remoteData) {
        throw Error('Provided remote is neither HTTP nor SSH');
    }
    var host = remoteData.host, owner = remoteData.owner, repo = remoteData.repo;
    var checkoutDir = Tmp.dirSync({ unsafeCleanup: true });
    var compareDir = Tmp.dirSync({ unsafeCleanup: true });
    var compareUrl = "https://" + host + "/" + owner + "/" + repo;
    try {
        // This will contain a new branch with 2 commits:
        // The first commit represents the current branch and the second one represents the changed one
        exports.Git(['init', compareDir.name]);
        // Create the commit for the current branch
        Fs.copySync(utils_1.Utils.cwd() + "/.git", checkoutDir.name + "/.git");
        exports.Git(['stash'], { cwd: checkoutDir.name });
        exports.Git(['fetch', remote], { cwd: checkoutDir.name });
        exports.Git(['checkout', "remotes/" + remote + "/" + branch], { cwd: checkoutDir.name });
        Fs.removeSync(checkoutDir.name + "/.git");
        Fs.moveSync(compareDir.name + "/.git", checkoutDir.name + "/.git");
        exports.Git(['add', '.'], { cwd: checkoutDir.name });
        exports.Git(['commit', '-m', "Current " + branch], { cwd: checkoutDir.name });
        Fs.moveSync(checkoutDir.name + "/.git", compareDir.name + "/.git");
        // Create the commit for the new branch
        Fs.copySync(utils_1.Utils.cwd() + "/.git", checkoutDir.name + "/.git");
        exports.Git(['stash'], { cwd: checkoutDir.name });
        exports.Git(['checkout', "refs/heads/" + branch], { cwd: checkoutDir.name });
        Fs.removeSync(checkoutDir.name + "/.git");
        Fs.moveSync(compareDir.name + "/.git", checkoutDir.name + "/.git");
        exports.Git(['add', '.'], { cwd: checkoutDir.name });
        try {
            exports.Git(['commit', '-m', "New " + branch], { cwd: checkoutDir.name });
        }
        catch (e) {
            throw Error('Remote and local branches are equal');
        }
        Fs.moveSync(checkoutDir.name + "/.git", compareDir.name + "/.git");
        // Prepare and push
        var currBranchHash = exports.Git(['rev-list', '--max-parents=0', 'HEAD'], { cwd: compareDir.name });
        var newBranchHash = exports.Git(['rev-parse', 'HEAD'], { cwd: compareDir.name });
        var compareBranch = "compare-" + currBranchHash.slice(0, 7) + "_" + newBranchHash.slice(0, 7);
        exports.Git(['checkout', '-b', compareBranch], { cwd: compareDir.name });
        exports.Git.print(['push', remoteUrl, compareBranch], { cwd: compareDir.name });
        exports.Git(['checkout', newBranchHash], { cwd: compareDir.name });
        exports.Git(['branch', '-D', compareBranch], { cwd: compareDir.name });
        exports.Git.print(['push', remoteUrl, ":refs/heads/" + compareBranch], { cwd: compareDir.name });
        compareUrl += "/compare/" + currBranchHash + ".." + newBranchHash;
    }
    finally {
        checkoutDir.removeCallback();
        compareDir.removeCallback();
    }
    return compareUrl;
}
// Push a tutorial based on the provided branch.
// e.g. given 'master' then 'master-history', 'master-root', 'master@0.1.0', etc, will be pushed.
// Note that everything will be pushed by FORCE and will override existing refs within the remote
function pushTutorial(remote, baseBranch) {
    var allBranches = git(['branch', '-l', '-a'])
        .split('\n')
        .map(function (b) { return b.split(/\*?\s+/).filter(Boolean)[0]; });
    var relatedBranches = allBranches.map(function (branch) {
        if (!branch) {
            return null;
        }
        var pathNodes = branch.split('/');
        if (pathNodes[0] === 'remotes') {
            if (pathNodes[1] !== remote) {
                return;
            }
            // `remotes/remote/branch` and `branch` are essentially the same and with duplication
            // an error will be thrown
        }
        else if (allBranches.includes("remotes/" + remote + "/" + branch)) {
            return;
        }
        var branchName = pathNodes.pop();
        if (branchName === baseBranch) {
            return branch;
        }
        if (branchName === baseBranch + "-history") {
            return branch;
        }
        if (branchName === baseBranch + "-root") {
            return branch;
        }
        if (new RegExp("^" + baseBranch + "-step\\d+$").test(branchName)) {
            return branch;
        }
    }).filter(Boolean);
    var relatedTags = git(['tag', '-l']).split('\n').map(function (tag) {
        if (!tag) {
            return null;
        }
        if (new RegExp("^" + baseBranch + "@(\\d+\\.\\d+\\.\\d+|next)$").test(tag)) {
            return tag;
        }
        if (new RegExp("^" + baseBranch + "@root@(\\d+\\.\\d+\\.\\d+|next)$").test(tag)) {
            return tag;
        }
        if (new RegExp("^" + baseBranch + "@step\\d+@(\\d+\\.\\d+\\.\\d+|next)$").test(tag)) {
            return tag;
        }
    }).filter(Boolean);
    var relatedBranchesNames = relatedBranches.map(function (b) { return b.split('/').pop(); });
    var deletedBranches = git(['ls-remote', '--heads', remote])
        .split('\n')
        .filter(Boolean)
        .map(function (line) { return line.split(/\s+/).pop(); })
        .filter(function (ref) { return !relatedBranchesNames.includes(ref.split('/').pop()); })
        .filter(function (ref) {
        var branch = ref.split('/').pop();
        return (branch === baseBranch ||
            branch === baseBranch + "-history" ||
            branch === baseBranch + "-root" ||
            new RegExp("^" + baseBranch + "-step\\d+$").test(branch));
    })
        // https://stackoverflow.com/questions/5480258/how-to-delete-a-remote-tag
        .map(function (ref) { return ":" + ref; });
    var deletedTags = git(['ls-remote', '--tags', remote])
        .split('\n')
        .filter(Boolean)
        .map(function (line) { return line.split(/\s+/).pop(); })
        .filter(function (ref) { return !relatedTags.includes(ref.split('/').pop()); })
        .filter(function (ref) {
        return new RegExp("^refs/tags/" + baseBranch + "@(\\d+\\.\\d+\\.\\d+|next)$").test(ref) ||
            new RegExp("^refs/tags/" + baseBranch + "@root@(\\d+\\.\\d+\\.\\d+|next)$").test(ref) ||
            new RegExp("^refs/tags/" + baseBranch + "@step\\d+@(\\d+\\.\\d+\\.\\d+|next)$").test(ref);
    })
        // https://stackoverflow.com/questions/5480258/how-to-delete-a-remote-tag
        .map(function (ref) { return ":" + ref; });
    var refs = relatedBranches.concat(relatedTags, deletedBranches, deletedTags);
    return gitPrint(['push', '-f', remote].concat(refs));
}
// Pull a tutorial based on the provided branch. e.g. given 'master' then 'master-history',
// 'master-root', 'master@0.1.0', etc, will be pulled.
function pullTutorial(remote, baseBranch) {
    var relatedBranches = [];
    var relatedTags = [];
    git(['ls-remote', '--tags', '--heads', remote]).split('\n').forEach(function (line) {
        if (!line) {
            return;
        }
        var _a = line.split(/\s+/), ref = _a[1];
        if (new RegExp("^refs/tags/" + baseBranch + "@(root|step-\\d+@)?(\\d+\\.\\d+\\.\\d+|next)$").test(ref)) {
            relatedTags.push(ref.split('/').slice(2).join('/'));
        }
        if (new RegExp("^refs/heads/" + baseBranch + "(-root|-history|-step\\d+)?$").test(ref)) {
            relatedBranches.push(ref.split('/').slice(2).join('/'));
        }
    });
    var refs = relatedBranches.concat(relatedTags);
    var activeBranchName = exports.Git.activeBranchName();
    try {
        var sha1 = exports.Git(['rev-parse', activeBranchName]);
        // Detach HEAD so we can change the reference of the branch
        exports.Git(['checkout', sha1]);
        // --tags flag will overwrite tags
        gitPrint(['fetch', '--tags', '-f', remote].concat(refs));
        // Make sure that all local branches track the right remote branches
        relatedBranches.forEach(function (branch) {
            try {
                exports.Git(['branch', '-D', branch]);
            }
            catch (e) {
                // Branch doesn't exist
            }
            exports.Git.print(['branch', '--track', branch, "remotes/" + remote + "/" + branch]);
        });
    }
    finally {
        // Get back to where we were, regardless of the outcome
        exports.Git(['checkout', activeBranchName]);
    }
}
// Used internally by tutorialStatus() to get the right step message
function getRefStep(ref) {
    if (ref === void 0) { ref = 'HEAD'; }
    if (getRootHash() === exports.Git(['rev-parse', ref])) {
        return 'root';
    }
    var match = exports.Git(['log', ref, '-1', '--format=%s']).match(/^Step (\d+(?:\.\d+)?)/);
    if (match) {
        return match[1];
    }
    return exports.Git(['rev-parse', '--short', ref]);
}
// Print edit status followed by git-status
function tutorialStatus(options) {
    if (options === void 0) { options = {}; }
    exports.Git.print(['status']);
    var instructions;
    position: if (isRebasing()) {
        var head = exports.Git(['rev-parse', 'HEAD']);
        var rebaseHead = exports.Git(['rev-parse', 'REBASE_HEAD']);
        var headStep = getRefStep('HEAD');
        if (head === rebaseHead) {
            console.log("\nEditing " + headStep);
            instructions = 'edit';
            break position;
        }
        var rebaseHeadStep = getRefStep('REBASE_HEAD');
        var isConflict = local_storage_1.localStorage.getItem('REBASE_NEW_STEP') !== headStep;
        if (isConflict) {
            console.log("\nSolving conflict between " + headStep + " and " + rebaseHeadStep);
            instructions = 'conflict';
            break position;
        }
        console.log("\nBranched out from " + rebaseHeadStep + " to " + headStep);
        instructions = 'edit';
    }
    switch (options.instruct && instructions) {
        case 'edit':
            console.log('\n' + utils_1.freeText("\n        To edit the current step, stage your changes and amend them:\n\n            $ git add xxx\n            $ git commit --amend\n\n        Feel free to push or pop steps:\n\n            $ tortilla step push/pop\n\n        Once you finish, continue the rebase and Tortilla will take care of the rest:\n\n            $ git rebase --continue\n\n        You can go back to re-edit previous steps at any point, but be noted that this will discard all your changes thus far:\n\n            $ tortilla step back\n\n        If for some reason, at any point you decide to quit, use the comand:\n\n            $ git rebase --abort\n      "));
        case 'conflict':
            console.log('\n' + utils_1.freeText("\n        Once you solved the conflict, stage your changes and continue the rebase.\n        DO NOT amend your changes, push or pop steps:\n\n            $ git add xxx\n            $ git rebase --continue\n\n        You can go back to re-edit previous steps at any point, but be noted that this will discard all your changes thus far:\n\n            $ tortilla step back\n\n        If for some reason, at any point you decide to quit, use the comand:\n\n            $ git rebase --abort\n      "));
    }
}
// The body of the git execution function, useful since we use the same logic both for
// exec and spawn
function gitBody(handler, argv, options) {
    options = __assign({ env: {} }, options);
    // Zeroing environment vars which might affect other executions
    options.env = __assign({ GIT_DIR: null, GIT_WORK_TREE: null }, options.env);
    return handler(argv, options);
}
// Tells if rebasing or not
function isRebasing(path) {
    if (path === void 0) { path = null; }
    var paths = path ? paths_1.resolveProject(path).git : paths_1.Paths.git;
    return utils_1.Utils.exists(paths.rebaseMerge) || utils_1.Utils.exists(paths.rebaseApply);
}
// Tells if cherry-picking or not
function isCherryPicking() {
    return utils_1.Utils.exists(paths_1.Paths.git.heads.cherryPick) || utils_1.Utils.exists(paths_1.Paths.git.heads.revert);
}
// Tells if going to amend or not
function gonnaAmend() {
    return utils_1.Utils.childProcessOf('git', ['commit', '--amend']);
}
// Tells if a tag exists or not
function tagExists(tag) {
    try {
        git(['rev-parse', tag]);
        return true;
    }
    catch (err) {
        return false;
    }
}
// Get the recent commit by the provided arguments. An offset can be specified which
// means that the recent commit from several times back can be fetched as well
function getRecentCommit(offset, argv, options, path) {
    if (path === void 0) { path = null; }
    if (offset instanceof Array) {
        options = argv;
        argv = offset;
        offset = 0;
    }
    else {
        argv = argv || [];
        offset = offset || 0;
    }
    var hash = typeof offset === 'string' ? offset : ("HEAD~" + offset);
    argv = ['log', hash, '-1'].concat(argv);
    return git(argv, path ? __assign({}, options, { cwd: path }) : options);
}
// Gets a list of the modified files reported by git matching the provided pattern.
// This includes untracked files, changed files and deleted files
function getStagedFiles(pattern) {
    var stagedFiles = git(['diff', '--name-only', '--cached'])
        .split('\n')
        .filter(Boolean);
    return utils_1.Utils.filterMatches(stagedFiles, pattern);
}
// Gets active branch name
function getActiveBranchName(path) {
    if (path === void 0) { path = null; }
    if (!isRebasing(path)) {
        return git(['rev-parse', '--abbrev-ref', 'HEAD'], path ? { cwd: path } : null);
    }
    // Getting a reference for the hash of which the rebase have started
    var branchHash = git(['reflog', '--format=%gd %gs'], path ? { cwd: path } : null)
        .split('\n')
        .filter(Boolean)
        .map(function (line) { return line.split(' '); })
        .map(function (split) { return [split.shift(), split.join(' ')]; })
        .find(function (_a) {
        var ref = _a[0], msg = _a[1];
        return msg.match(/^rebase -i \(start\)/);
    })
        .shift()
        .match(/^HEAD@\{(\d+)\}$/)
        .slice(1)
        .map(function (i) { return "HEAD@{" + ++i + "}"; })
        .map(function (ref) { return git(['rev-parse', ref]); })
        .pop();
    // Comparing the found hash to each of the branches' hashes
    return Fs.readdirSync(paths_1.Paths.git.refs.heads).find(function (branchName) {
        return git(['rev-parse', branchName], path ? { cwd: path } : null) === branchHash;
    });
}
// Gets the root hash of HEAD
function getRootHash(head, options) {
    if (head === void 0) { head = 'HEAD'; }
    if (options === void 0) { options = {}; }
    return git(['rev-list', '--max-parents=0', head], options);
}
function getRoot() {
    try {
        return git(['rev-parse', '--show-toplevel']);
        // Not a git project
    }
    catch (e) {
        return '';
    }
}
function edit(initialContent) {
    var editor = getEditor();
    var file = Tmp.fileSync();
    Fs.writeFileSync(file.name, initialContent);
    exec.print('sh', ['-c', editor + " " + file.name]);
    var content = Fs.readFileSync(file.name).toString();
    file.removeCallback();
    return content;
}
// https://github.com/git/git/blob/master/git-rebase--interactive.sh#L257
function getEditor() {
    var editor = process.env.GIT_EDITOR;
    if (!editor) {
        try {
            editor = git(['config', 'core.editor']);
        }
        catch (e) {
            // Ignore
        }
    }
    if (!editor) {
        try {
            editor = git(['var', 'GIT_EDITOR']);
        }
        catch (e) {
            // Ignore
        }
    }
    if (!editor) {
        throw Error('Git editor could not be found');
    }
    return editor;
}
// Commander will split equal signs e.g. `--format=%H` which is is not the desired
// behavior for git. This puts the everything back together when necessary
function normalizeArgv(argv) {
    argv = argv.slice();
    {
        var i = argv.indexOf('--format');
        if (i !== -1) {
            argv.splice(i, 2, "--format=" + argv[i + 1]);
        }
    }
    return argv;
}
function getRevisionIdFromObject(object) {
    return git(['rev-list', '-n', '1', object]);
}
exports.Git = utils_1.Utils.extend(git.bind(null), git, {
    pushTutorial: pushTutorial,
    pullTutorial: pullTutorial,
    reviewTutorial: reviewTutorial,
    tutorialStatus: tutorialStatus,
    conflict: conflict,
    rebasing: isRebasing,
    cherryPicking: isCherryPicking,
    gonnaAmend: gonnaAmend,
    tagExists: tagExists,
    recentCommit: getRecentCommit,
    stagedFiles: getStagedFiles,
    activeBranchName: getActiveBranchName,
    rootHash: getRootHash,
    root: getRoot,
    edit: edit,
    editor: getEditor,
    normalizeArgv: normalizeArgv,
    getRevisionIdFromObject: getRevisionIdFromObject,
});
//# sourceMappingURL=git.js.map