"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Fs = require("fs-extra");
var git_1 = require("../git");
var local_storage_1 = require("../local-storage");
var paths_1 = require("../paths");
var step_1 = require("../step");
var utils_1 = require("../utils");
/**
 Pre-commit git hook launches right before we commit changes. If an error was thrown
 the commit process will be aborted with the provided error message.
 */
function run() {
    if (process.env.TORTILLA_CHILD_PROCESS) {
        return;
    }
    if (!local_storage_1.localStorage.getItem('USE_STRICT')) {
        return;
    }
    // Prohibit regular commits
    if (!git_1.Git.gonnaAmend()) {
        throw Error('New commits are prohibited! Use `$ tortilla step push` instead');
    }
    if (!git_1.Git.rebasing()) {
        throw Error([
            'Changes are not allowed outside editing mode!',
            'Use `$ tortilla step edit` and then make your changes',
        ].join('\n'));
    }
    var stepMessage = step_1.Step.recentCommit('%s');
    var stepDescriptor = step_1.Step.descriptor(stepMessage);
    var isSuperStep = stepDescriptor && !stepDescriptor.number.split('.')[1];
    var stagedFiles;
    // If this is a super step only the appropriate manual file can be modified
    if (isSuperStep) {
        var tag_1 = "step" + stepDescriptor.number;
        var allowedFiles_1 = [
            ".tortilla/manuals/templates/" + tag_1 + ".tmpl",
            ".tortilla/manuals/views/" + tag_1 + ".md",
            '.tortilla/config.js',
            '.circleci/config.yml',
            'package.json',
        ];
        var localesDir = paths_1.Paths.manuals.templates + "/locales";
        if (utils_1.Utils.exists(localesDir)) {
            var locales = Fs.readdirSync(localesDir);
            allowedFiles_1 = allowedFiles_1.concat(locales.map(function (locale) { return ".tortilla/manuals/templates/locales/" + locale + "/" + tag_1 + ".tmpl"; }));
            allowedFiles_1 = allowedFiles_1.concat(locales.map(function (locale) { return ".tortilla/manuals/views/locales/" + locale + "/" + tag_1 + ".md"; }));
        }
        stagedFiles = git_1.Git.stagedFiles().filter(function (stagedFile) { return allowedFiles_1.indexOf(stagedFile) !== -1; });
        if (!stagedFiles.length) {
            var filesList = allowedFiles_1.map(function (file) { return "\u2022 " + file; }).join('\n');
            throw Error("Staged files must be one of:\n" + filesList);
        }
    }
    else if (stepDescriptor) {
        stagedFiles = git_1.Git.stagedFiles(/^\.tortilla\/manuals\//);
        if (stagedFiles.length) {
            throw Error('Step manual files can\'t be modified');
        }
    }
    stagedFiles = git_1.Git.stagedFiles(/^README.md/);
    if (stagedFiles.length) {
        throw Error([
            'README.md can\'t be modified.',
            'Run `$ tortilla step edit --root` and edit \'.tortilla/manuals/templates/root.md\' file instead',
        ].join('\n'));
    }
}
run();
//# sourceMappingURL=pre-commit.js.map