"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Handlebars = require("handlebars");
var ParseDiff = require("parse-diff");
var config_1 = require("../../config");
var git_1 = require("../../git");
var submodule_1 = require("../../submodule");
var translator_1 = require("../../translator");
var utils_1 = require("../../utils");
var index_1 = require("../index");
/**
  Renders step diff in a pretty markdown format. For example {{{ diffStep 1.1 }}}
  will render as:

  #### Step 1.1

  ##### Changed /path/to/file.js
  ```diff
  @@ -1,3 +1,3 @@
  +┊ ┊1┊foo
  -┊1┊ ┊bar
   ┊2┊2┊baz🚫↵
  ```

  We can also render this helper to be suitable for use in Medium (https://medium.com/)
  by setting the environment variable TORTILLA_RENDER_TARGET to 'medium':

  #### Step 1.1

  ##### Changed /path/to/file.js
  <pre>
  <i>@@ -1,3 +1,3 @@</i>
  <b>+┊ ┊1┊foo</b>
   ┊2┊2┊baz🚫↵
  </pre>

  Note: Removals won't be shown to reduce complexity, since medium can render only
  very simple stuff
 */
var t = translator_1.Translator.translate.bind(translator_1.Translator);
index_1.Renderer.registerHelper('diffStep', function (step, options) {
    var hash = options.hash;
    var pattern;
    // Will print diff of multiple specified files
    // e.g. files="foo/a, bar/b"
    if (hash.files) {
        pattern = new RegExp(hash.files.replace(/\s*,\s*/g, '|').replace(/\./g, '\\.'));
        // Will print diff of all possible files
    }
    else {
        pattern = /.*/;
    }
    var cwd = git_1.Git(['rev-parse', '--show-toplevel']);
    // In case a submodule was specified then all our git commands should be executed
    // from that module
    if (hash.module) {
        // Use the cloned repo that is used for development
        if (process.env.TORTILLA_SUBDEV) {
            cwd = submodule_1.Submodule.getCwd(hash.module);
        }
        else {
            cwd = cwd + "/" + hash.module;
        }
    }
    var stepData;
    if (step === 'root') {
        var rootHash = git_1.Git.rootHash('HEAD', { cwd: cwd });
        var message = git_1.Git(['log', rootHash, '-1', '--format=%s'], { cwd: cwd });
        // Simulating split result because it will be joined soon
        stepData = [rootHash].concat(message.split(' '));
    }
    else {
        stepData = git_1.Git.recentCommit([
            "--grep=^Step " + step + ":", '--format=%h %s',
        ], {
            cwd: cwd,
        }).split(' ')
            .filter(Boolean);
    }
    // In case step doesn't exist just render the error message.
    // It's better to have a silent error like this rather than a real one otherwise
    // the rebase process will skrew up very easily and we don't want that
    if (!hash.noTitle && !stepData.length) {
        return "#### " + t('step.commit.missing', { number: step });
    }
    var stepHash = stepData[0];
    var commitReference;
    // URL composition should be done in relation to the submodule
    if (hash.module) {
        commitReference = utils_1.Utils.tempCwd(function () {
            return index_1.Renderer.resolve('~/commit', stepHash);
        }, cwd);
    }
    else {
        commitReference = index_1.Renderer.resolve('~/commit', stepHash);
    }
    // Translate step message, if at all
    var stepMessage = index_1.Renderer.call('stepMessage', {
        commitMessage: stepData.slice(1).join(' '),
        module: hash.module,
    });
    var stepTitle;
    // If this is a relative path, we won't reference the commit
    if (commitReference.isRelative) {
        stepTitle = "#### " + stepMessage;
    }
    else {
        stepTitle = "#### [" + stepMessage + "](" + commitReference + ")";
    }
    var diff;
    if (step === 'root') {
        diff = git_1.Git(['show', stepHash, '--format='], { cwd: cwd });
    }
    else {
        diff = git_1.Git(['diff', stepHash + "^", stepHash], { cwd: cwd });
    }
    var blacklist = config_1.Config.getBlacklist();
    // Convert diff string to json format
    var files = ParseDiff(diff).filter(function (file) {
        // Filter files which match the given pattern
        return (file.from.match(pattern) || file.to.match(pattern)) &&
            (!blacklist || (!file.to.match(blacklist) && !file.from.match(blacklist)));
    });
    var mdDiffs = files
        .map(getMdDiff)
        .join('\n\n');
    if (hash.noTitle) {
        return mdDiffs;
    }
    return stepTitle + "\n\n" + mdDiffs;
}, {
    mdWrap: true,
});
// Gets all diff chunks in a markdown format for a single file
function getMdDiff(file) {
    var fileTitle;
    if (file.new) {
        fileTitle = "##### " + t('diff.added', { path: file.to });
    }
    else if (file.deleted) {
        fileTitle = "##### " + t('diff.deleted', { path: file.from });
    }
    else if (!file.chunks.length) {
        fileTitle = "##### " + t('diff.renamed', { from: file.from, to: file.to });
    }
    else {
        fileTitle = "##### " + t('diff.changed', { path: file.from });
    }
    fileTitle = fileTitle.replace(/__(\w+)__/g, '\\__$1__');
    var mdChunks = file.chunks
        .map(getMdChunk)
        .join('\n');
    return fileTitle + "\n" + mdChunks;
}
// Gets diff in a markdown format for a single chunk
function getMdChunk(chunk) {
    // Grab chunk data since it's followed by irrelevant content
    var chunkData = chunk.content.match(/^@@\s+\-(\d+),?(\d+)?\s+\+(\d+),?(\d+)?\s@@/)[0];
    var padLength = getPadLength(chunk.changes);
    var mdChanges = chunk.changes
        .map(getMdChange.bind(null, padLength))
        .join('\n')
        // Replace EOF flag with a pretty format and append it to the recent line
        .replace(/\n\\ No newline at end of file/g, '🚫↵');
    // Wrap changes with markdown 'diff'
    return ['```diff', chunkData, mdChanges, '```'].join('\n');
}
// Gets line in a markdown format for a single change
function getMdChange(padLength, change) {
    // No newline at end of file
    if (change.content[0] === '\\') {
        return change.content;
    }
    var addLineNum = '';
    var delLineNum = '';
    var sign = '';
    switch (change.type) {
        case 'add':
            sign = '+';
            addLineNum = change.ln;
            break;
        case 'del':
            sign = '-';
            delLineNum = change.ln;
            break;
        case 'normal':
            sign = ' ';
            addLineNum = change.ln2;
            delLineNum = change.ln1;
            break;
        default:
            break;
    }
    addLineNum = utils_1.Utils.pad(addLineNum, padLength);
    delLineNum = utils_1.Utils.pad(delLineNum, padLength);
    // Using content.slice(1) since we want to remove '-\+' prefixes
    return [sign, delLineNum, addLineNum, change.content.slice(1)].join('┊');
}
// Gets the pad length by the length of the max line number in changes
function getPadLength(changes) {
    var maxLineNumber = changes.reduce(function (max, change) { return (Math.max(max, change.ln || 0, change.ln1 || 0, change.ln2 || 0)); }, 1);
    return maxLineNumber.toString().length;
}
index_1.Renderer.registerTransformation('medium', 'diffStep', function (view) {
    var diffBlock = [
        '<i>╔══════╗</i>',
        '<i>║ diff ║</i>',
        '<i>╚══════╝</i>',
    ].join('\n');
    return view
        .split(/```diff\n|\n```(?!diff)/).map(function (chunk, index) {
        if (index % 2 === 0) {
            return chunk;
        }
        var content = Handlebars.escapeExpression(chunk)
            // Make diff changes (e.g. @@ -1,3 +1,3 @@) italic
            .replace(/^@.+$/m, diffBlock)
            // Remove removals
            .replace(/\n\-.+/g, '')
            // Bold additions
            .replace(/^(\+.+)$/mg, '<b>$&</b>');
        // Wrap with <pre> tag
        return "<pre>\n" + content + "\n</pre>";
    })
        .join('');
});
//# sourceMappingURL=diff-step.js.map