"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Fs = require("fs-extra");
var Handlebars = require("handlebars");
var Path = require("path");
var git_1 = require("../git");
var paths_1 = require("../paths");
var release_1 = require("../release");
var utils_1 = require("../utils");
/**
  A wrapper for Handlebars with several additions which are essential for Tortilla
 */
// Creating a new instance of handlebars which will then be merged with the module
var handlebars = Handlebars.create();
// Keep original handlers since these methods are gonna be overriden
var superRegisterHelper = handlebars.registerHelper.bind(handlebars);
var superRegisterPartial = handlebars.registerPartial.bind(handlebars);
// Used to store registered transformations
var transformations = {};
// Cache for templates which were already compiled
var cache = {};
// Read the provided file, render it, and overwrite it. Use with caution!
function overwriteTemplateFile(templatePath, scope) {
    templatePath = resolveTemplatePath(templatePath);
    var view = renderTemplateFile(templatePath, scope);
    return Fs.writeFileSync(templatePath, view);
}
// Read provided file and render its template. Note that the default path would
// be tortilla's template dir, so specifying a file name would be ok as well
function renderTemplateFile(templatePath, scope) {
    templatePath = resolveTemplatePath(templatePath);
    if (process.env.TORTILLA_CACHE_DISABLED || !cache[templatePath]) {
        var templateContent = Fs.readFileSync(templatePath, 'utf8');
        cache[templatePath] = handlebars.compile(templateContent);
    }
    var template = cache[templatePath];
    return renderTemplate(template, scope);
}
// Render provided template
function renderTemplate(template, scope) {
    // Template can either be a string or a compiled template object
    if (typeof template === 'string') {
        template = handlebars.compile(template);
    }
    scope = scope || {};
    var viewDir;
    if (scope.viewPath) {
        // Relative path of view dir
        // e.g. manuals/views
        viewDir = Path.dirname(scope.viewPath);
    }
    var oldResolve = handlebars.resolve;
    try {
        // Set the view file for the resolve utility. If no view path was provided, the
        // resolve function below still won't work
        handlebars.resolve = resolvePath.bind(null, viewDir);
        // Remove trailing white-space
        return template(scope).replace(/ *\n/g, '\n');
    }
    finally {
        // Either if an error was thrown or not, unbind it
        handlebars.resolve = oldResolve;
    }
}
// Returns a template path relative to tortilla with an '.tmpl' extension
function resolveTemplatePath(templatePath) {
    if (templatePath.indexOf('.tmpl') === -1) {
        templatePath += '.tmpl';
    }
    // User defined templates
    var relativeTemplatePath = Path.resolve(paths_1.Paths.manuals.templates, templatePath);
    if (utils_1.Utils.exists(relativeTemplatePath)) {
        return relativeTemplatePath;
    }
    // Tortilla defined templates
    return Path.resolve(paths_1.Paths.tortilla.renderer.templates, templatePath);
}
// Register a new helper. Registered helpers will be wrapped with a
// [{]: <helper> (name ...args) [}]: #
function registerHelper(name, helper, options) {
    options = options || {};
    var wrappedHelper = function () {
        var oldCall = handlebars.call;
        var out;
        try {
            // Bind the call method to the current context
            handlebars.call = callHelper.bind(this);
            out = helper.apply(this, arguments);
        }
        finally { // Fallback
            // Restore method to its original
            handlebars.call = oldCall;
        }
        if (typeof out !== 'string' &&
            !(out instanceof String)) {
            throw Error([
                'Template helper', name, 'must return a string!',
                'Instead it returned', out,
            ].join(' '));
        }
        var target = process.env.TORTILLA_RENDER_TARGET;
        var args = [].slice.call(arguments);
        // Transform helper output
        var transformation = transformations[target] && transformations[target][name];
        if (transformation) {
            out = transformation.apply(void 0, [out].concat(args));
        }
        // Wrap helper output
        if (options.mdWrap) {
            out = mdWrapComponent('helper', name, args, out);
        }
        return out;
    };
    superRegisterHelper(name, wrappedHelper);
}
// Register a new partial. Registered partials will be wrapped with a
// [{]: <partial> (name) [}]: #
function registerPartial(name, partial, options) {
    options = options || {};
    // Wrap partial template
    if (options.mdWrap) {
        partial = mdWrapComponent('partial', name, partial);
    }
    return superRegisterPartial(name, partial);
}
// Register a new transformation which will take effect on rendered helpers. This is
// useful when setting the TORTILLA_RENDER_TARGET variable, so we can make additional
// adjustments for custom targets. For now this is NOT part of the official API and
// is used only for development purposes
function registerTransformation(targetName, helperName, transformation) {
    if (!transformations[targetName]) {
        transformations[targetName] = {};
    }
    transformations[targetName][helperName] = transformation;
}
// Returns content wrapped by component notations. Mostly useful if we want to detect
// components in the view later on using external softwares later on.
// e.g. https://github.com/Urigo/angular-meteor-docs/blob/master/src/app/tutorials/
// improve-code-resolver.ts#L24
function mdWrapComponent(type, name, args, content) {
    var hash = {};
    if (typeof content !== 'string') {
        content = args;
        args = [];
    }
    if (args[args.length - 1] instanceof Object) {
        hash = args.pop().hash;
    }
    // Stringify arguments
    var params = args.map(function (param) { return typeof param === 'string' ? "\"" + param + "\"" : param; }).join(' ');
    hash = stringifyHash(hash);
    // Concat all stringified arguments
    args = [name, params, hash]
        // Get rid of empty strings
        .filter(Boolean)
        .join(' ');
    return "[{]: <" + type + "> (" + utils_1.Utils.escapeBrackets(args) + ")\n\n" + content + "\n\n[}]: #";
}
// Takes a helper hash and stringifying it
// e.g. { foo: '1', bar: 2 } -> foo="1" bar=2
function stringifyHash(hash) {
    return Object.keys(hash).map(function (key) {
        var value = hash[key];
        if (typeof value === 'string') {
            value = "\"" + value + "\"";
        }
        return key + "=" + value;
    }).join(' ');
}
// Calls a template helper with the provided context and arguments
function callHelper(methodName) {
    var args = [].slice.call(arguments, 1);
    var options = args.pop();
    // Simulate call from template
    if (options instanceof Object) {
        options = { hash: options };
    }
    args.push(options);
    return handlebars.helpers[methodName].apply(this, args);
}
// Takes a bunch of paths and resolved them relatively to the current rendered view
function resolvePath( /* reserved path, user defined path */) {
    var paths = [].slice.call(arguments);
    // A default path that the host's markdown renderer will know how to resolve by its own
    var defaultPath = paths.slice(1).join('/');
    /* tslint:disable-next-line */
    defaultPath = new String(defaultPath);
    // The 'isRelative' flag can be used later on to determine if this is an absolute path
    // or a relative path
    defaultPath.isRelative = true;
    var cwd = paths.shift();
    // If function is unbound, return default path
    if (typeof cwd !== 'string') {
        return defaultPath;
    }
    var repository = Fs.readJsonSync(paths_1.Paths.npm.package).repository;
    var repositoryUrl = typeof repository === 'object' ? repository.url : repository;
    // If no repository was defined, or
    // repository type is not git, or
    // no repository url is defined, return default path
    if (!repositoryUrl) {
        return defaultPath;
    }
    repositoryUrl = repositoryUrl
        // Remove .git postfix
        .replace(/.git$/, '')
        // Remove USERNAME@githost.com prefix
        .replace(/[\w]+@/g, '');
    var releaseTag;
    // First try to see if HEAD is referencing a release tag already. This is necessary
    // if we're running executing the current template helper for a submodule
    try {
        releaseTag = git_1.Git(['describe', '--tags', '--exact-match', 'HEAD']);
        // If not, manually compose it. Necessary for the main git-module
    }
    catch (e) {
        var currentRelease = release_1.Release.format(release_1.Release.current());
        // Any release is yet to exist
        if (currentRelease === '0.0.0') {
            return defaultPath;
        }
        releaseTag = git_1.Git.activeBranchName() + "@" + currentRelease;
    }
    // Compose branch path for current release tree
    // e.g. github.com/Urigo/Ionic2CLI-Meteor-Whatsapp/tree/master@0.0.1
    var branchUrl = [repositoryUrl, 'tree', releaseTag].join('\/');
    var protocol = (branchUrl.match(/^.+\:\/\//) || [''])[0];
    // Using a / so we can use it with the 'path' module
    var branchPath = "/" + branchUrl.substr(protocol.length);
    // If we use tilde (~) at the beginning of the path, we will be referenced to the
    // repo's root URL. This is useful when we want to compose links which are
    // completely disconnected from the current state, like commits, issues and PRs
    var resolved = paths.map(function (path) { return path.replace(/\~/g, Path.resolve(branchPath, '../..')); });
    // The view dir URL
    // e.g. github.com/DAB0mB/radial-snake/tree/master@0.1.5/.tortilla/manuals/views
    resolved.unshift(Path.join(branchPath, Path.relative(utils_1.Utils.cwd(), Path.resolve(utils_1.Utils.cwd(), cwd))));
    resolved = Path.resolve.apply(Path, resolved);
    // Concatenating the protocol back after final path has been formed
    resolved = protocol + resolved.substr(1);
    return resolved;
}
exports.Renderer = utils_1.Utils.extend(handlebars, {
    overwriteTemplateFile: overwriteTemplateFile,
    renderTemplateFile: renderTemplateFile,
    renderTemplate: renderTemplate,
    registerHelper: registerHelper,
    registerPartial: registerPartial,
    registerTransformation: registerTransformation,
    // This should be set whenever we're in a helper scope
    call: callHelper,
    // Should be bound by the `renderTemplate` method
    resolve: resolvePath.bind(null, null),
});
// Built-in helpers and partials
require("./helpers/comment");
require("./helpers/diff-step");
require("./helpers/nav-step");
require("./helpers/resolve-path");
require("./helpers/step-message");
require("./helpers/translate");
//# sourceMappingURL=index.js.map